import sys
from inspect import signature

try:  # Python 2.x
    from StringIO import StringIO
except ImportError:  # Python 3.x
    from io import StringIO

from django.core.management import BaseCommand
from django.conf import settings
from importlib import import_module


class Command(BaseCommand):
    TAB = ' ' * 4
    GREEN = '\033[32m'
    BOLD_GREEN = '\033[1;32m'
    BOLD_YELLOW = '\033[1;33m'
    BOLD_BLUE = '\033[1;34m'
    RESET = '\033[0m'

    def handle(self, *args, **kwargs):
        for service_string, service, methods in self.get_services():
            self.print_service_info(service_string, service, methods)

    def get_services(self):
        for service_string in settings.ISC['services']:
            service = self._import_service(service_string)
            yield (service_string, service, list(self._get_service_methods(service)))

    def _import_service(self, service_string):
        module_name, _, class_name = service_string.rpartition('.')
        module = import_module(module_name)
        return getattr(module, class_name)

    def _get_service_methods(self, service):
        for attr in dir(service):
            fn = getattr(service, attr)
            if hasattr(fn, '__exposed__'):
                yield fn

    def print_service_info(self, service_string, service, methods):
        service_module, _, service_class = service_string.rpartition('.')
        sys.stdout.write(Command.GREEN + service_module + '.' + Command.BOLD_GREEN + service_class + Command.RESET + ': ')
        sys.stdout.write(Command.BOLD_BLUE + service.name + Command.RESET + '\n')
        for method in methods:
            self.print_method_info(method)

    def print_method_info(self, method):
        fn_str = Command.BOLD_YELLOW + method.__name__ + Command.RESET + str(signature(method))
        sys.stdout.write(
            Command.TAB + '- ' + fn_str + '\n'
        )



import subprocess
import sys
try:
    # Python 3.x
    from shutil import which
except ImportError:
    # Python 2.x
    from distutils.spawn import find_executable as which

try:
    from IPython import start_ipython
    ipython = True
except:
    ipython = False

from isc.client import Client

from django.core.management import BaseCommand
from django.conf import settings
from importlib import import_module


class Command(BaseCommand):
    CMD = \
        'from isc.client import Client;' \
        'from threading import Event;' \
        'rpc = Client(host="{}");' \
        'e = Event();' \
        'rpc._consumer.on_connect += e.set;' \
        'rpc.start();' \
        'e.wait();' \
        'rpc._consumer.on_connect -= e.set;' \
        'del e;' \
        'print(\'\\n\\033[1;34mINFO: You now have the global "rpc" variable.\\033[0m\\n\')'

    def handle(self, *args, **kwargs):
        if not hasattr(settings, 'ISC'):
            sys.stderr.write('Django config is missing "ISC" configuration.')
            sys.exit(1)
        if not settings.ISC['url']:
            sys.stderr.write('Django config for ISC is missing "url" key.')
            sys.exit(1)
        args = ['-i', '-c', Command.CMD.format(settings.ISC['url'])]
        if ipython:
            sys.argv = [sys.executable] + args
            start_ipython()
        else:
            ps = subprocess.Popen(
                [sys.executable] + args,
                stdin=sys.stdin,
                stdout=sys.stdout,
                stderr=sys.stderr
            )
            ps.communicate()
            sys.exit(ps.returncode)




from isc.client import FutureResult, RemoteException

try:
    from unittest.mock import patch
except:
    import mock


class FakeInvocationProxy(object):
    def __init__(self, definitions):
        self.definitions = definitions

    def __call__(self, service, method, *args, **kwargs):
        def_key = '{}.{}'.format(service, method)
        assert def_key in self.definitions, 'Your code tried to call "{}" which is not present in `patch_isc` definition.'.format(
            def_key
        )
        def_value = self.definitions[def_key]
        print('FakeInvocationProxy: calling {}.{}(*{}, **{})'.format(
            service,
            method,
            repr(args),
            repr(kwargs)
        ))
        future = FutureResult('{}.{}'.format(service, method))
        try:
            if callable(def_value):
                result = def_value(*args, **kwargs)
            else:
                result = def_value
            future.resolve(result)
        except Exception as e:
            future.reject(RemoteException(str(e)))
        return future


def patch_isc(definitions):
    """
    Patch ISC return for specified methods.
    `definitions` should be a dictionary where each key
    is a string in "service.method" form and each value
    is a predefined return value or a callable.
    """
    def decorator(fn):
        def wrapper(*args, **kwargs):
            with patch(
                'isc.client.Client.invoke_async',
                side_effect=FakeInvocationProxy(definitions)
            ):
                return fn(*args, **kwargs)
        return wrapper
    return decorator



import pickle
import json
from uuid import UUID
from dateutil import parser
from datetime import datetime


class CodecException(Exception):
    pass


class AbstractCodec(object):
    """
    Abstract base class for implementing codecs.

    "Codec" is a class that tells ISC how to encode & decode message payloads.

    Server can support multiple codecs while client can use only one at a time.

    Default codec is :class:`.PickleCodec`. You can implement your own codec
    by extending :class:`.AbstractCodec` and overriding its methods.

    Important: Don't forget that both client and server should have the codec installed!
    """
    def encode(self, message):  # pragma: no cover
        """
        Called when a message needs to be serialized
        for sending over AMQP channel.
        """
        raise NotImplementedError()

    def decode(self, payload):  # pragma: no cover
        """
        Called when a message needs to be serialized
        for sending over AMQP channel.
        """
        raise NotImplementedError()


class PickleCodec(AbstractCodec):
    """
    Pickle codec implementation.
    """
    content_type = 'pickle'

    def encode(self, message):
        return pickle.dumps(message, protocol=2)

    def decode(self, payload):
        return pickle.loads(payload)


class JSONCodec(AbstractCodec):
    """
    JSON codec implementation.
    """
    content_type = 'json'

    def encode(self, message):
        return json.dumps(message)

    def decode(self, payload):
        return json.loads(payload)


class TypedJSONCodec(AbstractCodec):
    """
    JSON codec implementation with support of timestamps & UUIDs.
    """
    content_type = 'typedjson'

    def encode(self, message):
        encoder = json.JSONEncoder(default=self._encode_object)
        return encoder.encode(message)

    def _encode_object(self, v):
        if isinstance(v, datetime):
            return dict(__object_type='datetime', __object_value=v.isoformat())
        elif isinstance(v, UUID):
            return dict(__object_type='uuid', __object_value=v.hex)
        else:
            raise CodecException('Don\'t know how to serialize {}'.format(repr(v)))

    def decode(self, payload):
        return json.JSONDecoder(object_hook=self._decode_object).decode(payload)

    def _decode_object(self, v):
        try:
            obj_type = v['__object_type']
            obj_value = v['__object_value']
        except:
            return v

        if obj_type == 'datetime':
            return parser.parse(obj_value)
        elif obj_type == 'uuid':
            return UUID(obj_value)
        else:
            raise CodecException('Unknown type: {}'.format(obj_type))

